"""
 Copyright (c) 2005-2006 Guido Wesdorp. All rights reserved.
 This software is distributed under the terms of the Templess
 License. See LICENSE.txt for license text.
 E-mail: johnny@johnnydebris.net

 A very compact, XML based templating system. It has only 5 different 
 directives, and doesn't allow any logic inside the templates, but
 instead expects the application to provide a dict with all the data (as
 strings or XML nodes) completely prepared. This makes it quite restrictive,
 and harder to seperately work on design and logic (the template has to very
 strictly adhere to the format of the dict), but makes it an environment
 extremely suitable for developers (since all development is done from code
 rather than from the template, and they don't have to learn a new, quirky 
 templating language).
 
"""

import nanosax

__appname__ = 'templess'
__version__ = '0.2'
__author__ = 'Guido Wesdorp <johnny@johnnydebris.net>'
__last_modified_date__ = \
    '$Date: 2006-10-06 13:20:37 +0200 (Fri, 06 Oct 2006) $'
__last_author__ = '$Author: johnny $'
__revision__ = '$Revision: 86 $'
__footer__ = '%s v%s, (c) %s 2005' % (__appname__, __version__, __author__)

XMLNS = 'http://johnnydebris.net/xmlns/templess'

# some classes for templess users
class xmlstring(unicode):
    """wrapper around unicode to mark a string as XML

        when Templess encounters this object, it will not escape entities
    """

def iterwrapper(iterable):
    for i in iterable:
        yield objectcontext(i)

class objectcontext(object):
    """wrapper around objects to allow Templess to traverse attributes

        this is similar to the nastyness Zope exposes to its users in TAL,
        when Templess retrieves a key from the context, if the context is
        an objectcontext wrapping an object, instead of just doing a 
        __getitem__ the following will be done:

        - first a __getitem__ is tried, then a __getattr_, if the value is
          still not found a KeyError is raised

        - before the value is returned, a check is done whether it's callable,
          and if so it's called (without arguments)
    """
    # XXX do we want to allow arguments in something like "@key,key" form or
    # something scary like that? this is nasty already anyway ;)
    def __init__(self, o):
        self.object = o

    __marker__ = []
    def __getitem__(self, name):
        try:
            ret = self.object[name]
        except (TypeError, KeyError):
            try:
                ret = self.object.__dict__[name]
            except KeyError:
                raise KeyError, name
        if callable(ret):
            ret = ret()
        return self.wrap(ret)

    def __str__(self):
        return getattr(self.object, '__str__', self.object.__repr__)()

    def __repr__(self):
        return '<objectcontext for "%s">' % (self.object.__class__.__name__,)

    def __getattr__(self, name):
        # XXX this is _so_ nasty... :|
        ret = self.wrap(getattr(self.object, name))

    def wrap(self, obj):
        if is_iterable_not_string(obj):
            return iterwrapper(obj)
        return objectcontext(obj)

    def __hasattr__(self, name):
        return hasattr(self.object, name)

# helper methods
def is_iterable_not_string(v):
    """determine if something is iterable but not a string"""
    if (type(v) in [str, unicode, xmlstring, objectcontext]
            or isinstance(v, node)):
        return False
    try:
        iter(v)
    except TypeError:
        pass
    else:
        return True
    return False

# tree to string conversion stuff
def entitize(s):
    """entitize a string so it can be used in XML content and attrs"""
    return s.replace('&', '&amp;').replace('"', '&quot;').replace(
                '<', '&lt;').replace('>', '&gt;').replace("'", '&apos;')

# tree stuff
class node(object):
    """node base
    
        very specific to Templess, the nodes contain as little functionality
        as possible
    """

    def find(self, name):
        return []

class elnode(list, node):
    """XML element"""

    def __init__(self, name, attrs, parent, charset='UTF-8'):
        self.name = name
        self.attrs = dict(attrs)
        self.parent = parent
        self.charset = charset
        if parent is not None:
            parent.append(self)

    def unicode(self):
        """return a string (XML) representation of ourselves (unrendered!)"""
        ret = [self._start_node_start(self.attrs)]
        if not len(self):
            ret.append(u' />')
        else:
            ret.append(u'>')
            for child in self:
                ret.append(child.unicode())
            ret.append(self._end_node())
        return ''.join(ret)

    def generate(self, context):
        """returns self as a generator (yielding unicode strings)"""
        issingle = not len(self)
        for chunk in self._start_node(self.attrs, issingle):
            yield chunk
        for child in self:
            for chunk in child.generate(context):
                yield chunk
        if len(self):
            yield self._end_node()

    def __repr__(self):
        return '<%s "%s">' % (self.__class__.__name__, self.name)

    def find(self, name):
        if self.name == name:
            yield self
        for child in self:
            for found in child.find(name):
                yield found

    def _strconvert(self, s):
        if isinstance(s, node):
            return s
        elif isinstance(s, xmlstring):
            return unicode(s)
        elif isinstance(s, str):
            s = unicode(s, self.charset)
        elif not isinstance(s, unicode):
            s = str(s)
        return entitize(s)

    def _start_node(self, attrs, single):
        yield self._start_node_start(attrs)
        if single:
            yield u' />'
        else:
            yield u'>'

    def _start_node_start(self, attrs):
        return u'<%s%s' % (self.name, self._serialize_attrs(attrs))

    def _end_node(self):
        return u'</%s>' % (self.name,)

    def _serialize_attrs(self, attrs):
        """return the attributes as a string"""
        if not len(attrs):
            return ''
        items = sorted(attrs.items())
        return ' %s' % ' '.join(['%s="%s"' % (k, self._strconvert(v))
                                    for (k, v) in items])

class templessnode(elnode):
    """templess element node

        has a special method 'convert' that makes it conver itself to a
        normal elnode by processing all templess directives (returns a new
        node, doesn't process in-place)
    """

    def __init__(self, name, attrs, parent, t_prefix, charset='UTF-8'):
        self.name = name
        self.attrs = dict(attrs)
        self.parent = parent
        self.t_prefix = t_prefix
        self.charset = charset
        if parent is not None:
            parent.append(self)

    def convert(self, context, parent=None):
        """process the templess directives
        
            returns a copy of self, or a list of copies of self (in case of
            t:content), can return unexpected results in case of t:replace,
            use only on root nodes of documents in production situation
            (iow: if you want to convert a tree, always make sure there's no
            t:repeat on the node you call convert() on)

            context is a context dict (see docs), parent is an internal helper
            argument
        """
        attrs = self.attrs.copy()
        # first see if we need to be rendered at all
        cont = self._handle_cond(attrs, context)
        if not cont:
            return ''
        replacekey = self._get_replace(attrs)
        contentkey = self._get_content(attrs)
        attrs = self._process_attr(attrs, context)

        # find out what context we need to use for rendering our children, or
        # what data we need to interpolate
        cvalue = None
        if replacekey:
            cvalue = context[replacekey]
        elif contentkey:
            cvalue = context[contentkey]

        # render our children
        if cvalue is None:
            retnode = parent
            if not replacekey:
                retnode = elnode(self.name, attrs, parent, self.charset)
            for child in self:
                child.convert(context, retnode)
            return retnode
        elif is_iterable_not_string(cvalue):
            ret = []
            for ccontext in cvalue:
                if (isinstance(ccontext, dict) or 
                        isinstance(ccontext, objectcontext)):
                    newparent = parent
                    if not replacekey:
                        # create a similar node to retnode
                        newparent = elnode(self.name, attrs, parent,
                                            self.charset)
                    for child in self:
                        child.convert(ccontext, newparent)
                        # return the last of the new nodes
                    ret.append(newparent)
                else:
                    newparent = parent
                    if not replacekey:
                        newparent = elnode(self.name, attrs, parent,
                                            self.charset)
                    if isinstance(ccontext, node):
                        if ccontext.parent:
                            ccontext.parent.remove(ccontext)
                        ccontext.parent = newparent
                        newparent.append(ccontext)
                    else:
                        textnode(self._strconvert(ccontext), newparent,
                                    self.charset)
                    ret.append(newparent)
            return ret
        else:
            retnode = parent
            if not replacekey:
                retnode = elnode(self.name, attrs, parent, self.charset)
            if isinstance(cvalue, node):
                if cvalue.parent:
                    cvalue.parent.remove(cvalue)
                cvalue.parent = retnode
                retnode.append(cvalue)
            else:
                textnode(self._strconvert(cvalue), retnode, self.charset)
            return retnode

    def _handle_cond(self, attrs, context):
        """handle the 't:cond' and 't:not' directives
        
            returns False if there's a t:cond or t:not that doesn't allow 
            rendering, True otherwise
        """
        attrkey = '%s:cond' % (self.t_prefix,)
        cond = attrs.pop(attrkey, False)
        if cond:
            return not not context[cond]
        attrkey = '%s:not' % (self.t_prefix,)
        cnot = attrs.pop(attrkey, False)
        if cnot:
            return not context[cnot]
        return True

    def _process_attr(self, attrs, context):
        attrkey = '%s:attr' % (self.t_prefix,)
        strvalue = attrs.pop(attrkey, False)
        if strvalue:
            pairs = [v.strip() for v in strvalue.split(';')]
            for pair in pairs:
                k, v = pair.split(' ')
                value = context[v]
                if not value:
                    attrs.pop(k, None)
                else:
                    attrs[k] = value
        return attrs

    def _get_replace(self, attrs):
        attrkey = '%s:replace' % (self.t_prefix,)
        replace = attrs.pop(attrkey, False)
        return replace

    def _get_content(self, attrs):
        attrkey = '%s:content' % (self.t_prefix,)
        content = attrs.pop(attrkey, False)
        return content

class textnode(node):
    """text element"""

    def __init__(self, text, parent, charset='UTF-8'):
        if not isinstance(text, unicode):
            text = unicode(text, charset)
        self.text = text
        self.parent = parent
        self.charset = charset
        if parent is not None:
            parent.append(self)

    def unicode(self):
        return self.text

    def generate(self, context):
        yield self.unicode()

    def convert(self, context, parent):
        return self.__class__(self.text, parent)

    def __repr__(self):
        reprtext = self.text[10:]
        if len(reprtext) < len(self.text):
            reprtext += '...'
        return '<%s "%s">' % (self.__class__.__name__, reprtext)

class cdatanode(textnode):
    """cdata node"""

    def unicode(self):
        return '<![CDATA[%s]]>' % (self.text,)

class commentnode(textnode):
    """comment node"""

    def unicode(self):
        return '<!--%s-->' % (self.text,)

# nanosax handler
class treebuilderhandler(nanosax.nshandler):
    """nanosax handler to convert XML to a node tree"""
    def __init__(self, charset):
        self.charset = charset

    def startdoc(self):
        self.current = None
        self.root = None
        self.templess_prefix = None

    def enddoc(self):
        pass

    def startel(self, name, attrs):
        for key, value in attrs.items():
            if key.startswith('xmlns') and value == XMLNS:
                if not ':' in key:
                    self.templess_prefix = ''
                else:
                    self.templess_prefix = key.split(':')[1]
        self.current = templessnode(name, attrs,
                        self.current, self.templess_prefix,
                                self.charset)
        if self.root is None:
            self.root = self.current

    def endel(self, name):
        self.current = self.current.parent

    def text(self, text):
        textnode(text, self.current, self.charset)

    def comment(self, text):
        commentnode(text, self.current, self.charset)

    def cdata(self, text):
        cdatanode(text, self.current, self.charset)

class template(object):
    """a Templess template

        call 'unicode()' to get a unicode string, 'generate()' to get a 
        generator that yields bits of string, and 'render()' to get a node
    """
    def __init__(self, data, charset='UTF-8'):
        self.charset = charset
        if hasattr(data, 'read'):
            data = data.read()
        if charset != 'UTF-8':
            data = unicode(data, charset)
        if isinstance(data, unicode):
            data = data.encode('UTF-8')
        handler = treebuilderhandler(charset)
        parser = nanosax.nsparser(handler)
        parser.parse(data)
        self.tree = handler.root

    def render_to_string(self, context, charset='UTF-8'):
        """returns a complete string with the rendered template"""
        # not preferred anymore...
        return self.unicode(context).encode(charset)

    def unicode(self, context):
        """returns a unicode rendering of the template"""
        node = self.tree.convert(context)
        return node.unicode()

    def generate(self, context):
        """returns a generator that generates bits of string on demand"""
        node = self.tree.convert(context)
        return node.generate()

    def render(self, context):
        """render the tree to a 'normal' node
        
            processes all templess directives and returns the resulting tree
        """
        node = self.tree.convert(context)
        return node

class cgitemplate:
    """simple cgi engine for serving templess templates"""
    def __init__(self, template, content_type='text/html; charset=UTF-8',
                    error_template='templates/cgierror.html', headers={}):
        self.template = template
        self.content_type = content_type
        self.error_template = error_template
        self.headers = headers

    def render(self, context):
        """render and output the document

            this also prints the Content-Type header, if you want it to print
            other headers use the headers arg of the constructor
        """
        if not context.has_key('footer'):
            context['footer'] = __footer__
        try:
            fp = open(self.template)
            try:
                t = template(fp)
                ret = t.render_to_string(context)
            finally:
                fp.close()
            assert type(ret) == str
            self.print_output(self.content_type, ret)
        except:
            import sys
            exc, e, tb = sys.exc_info()
            ret = self.handle_exception(exc, e, tb)

    def handle_exception(self, exc, e, tb):
        """internal method that generates and prints an exception page"""
        from traceback import format_tb
        tblist = format_tb(tb)
        fp = open(self.error_template)
        try:
            t = template(fp)
            context = {
                'exception': str(exc),
                'message': str(e),
                'traceback': '\n'.join(tblist),
                'tblist': [{'line': str(l)} for l in tblist],
                'footer': __footer__,
            }
            ret = t.render_to_string(context)
        finally:
            fp.close()
        self.print_output('text/html; charset=UTF-8', ret)

    def print_output(self, content_type, body):
        """actually print the output"""
        print 'Content-Type: %s' % content_type
        for kv in self.headers.items():
            print '%s: %s' % kv
        print
        print body

if __name__ == '__main__':
    # some test code
    import sys
    if len(sys.argv) != 2:
        print 'usage: %s <xmlfile>' % (sys.argv[0],)
        sys.exit()
    xml = open(sys.argv[1]).read()
    handler = treebuilderhandler('UTF-8')
    parser = nanosax.nsparser(handler)
    parser.parse(xml)
    print handler.root



from __future__ import division

__author__ = 'Roberto De Almeida <rob@pydap.org>'

import re
import bisect
from cStringIO import StringIO

from paste.request import construct_url, parse_dict_querystring
from paste.httpexceptions import HTTPBadRequest
import Image, ImageDraw, ImageColor
import numpy
from numpy import ma

from dap.lib import __dap__
from dap.dtypes import *

from dap.responses.wms import templess
from dap.responses.wms.palette import jet


def build(self, constraints=None):
    # Parse query string.
    query = parse_dict_querystring(self.environ)

    # Grab full dataset.
    dataset = self._parseconstraints(constraints)
    
    # Get request type.
    request = query.get('REQUEST', 'GetMap')
    if request == 'GetCapabilities':
        return _capabilities(dataset, self.description, self.environ)
    elif request != 'GetMap':
        raise HTTPBadRequest('You need to specify a REQUEST type')

    # Filter sequences by constraint. This is importante because it avoids
    # building a complete sequence from the database just to filter it later.
    seqs = [var for var in dataset if isinstance(var, SequenceType)]
    for seq in seqs:
        # Grab lat & lon variables from the sequence.
        lat = lon = None
        for var in seq.walk():
            if getattr(var, "axis", None) == 'X': lon = var
            if getattr(var, "axis", None) == 'Y': lat = var

        # Get requested boundaries.
        bbox = [float(v) for v in query.get('BBOX', '-180,-90,180,90').split(',')]
        minlon, minlat, maxlon, maxlat = bbox

        # Put coordinates in same range as the data. For this the sequence
        # should have a ``lon_range`` attributes (as the dapper spec requires).
        try: lon_range = dataset.attributes['NC_GLOBAL']['lon_range']
        except: lon_range = -360.0, 360.0
        query['lon_range'] = lon_range
        minlon = lon_range[0] - (lon_range[0] % 360.0) + (minlon % 360.0)
        maxlon = lon_range[0] - (lon_range[0] % 360.0) + (maxlon % 360.0)
        
        # Filter data accordingly. We can only filter data when ``minlon`` is 
        # less than ``maxlon`` -- this is not always the case because longitudes
        # wrap. Since the DAP does not allow OR in the constraint expression
        # we can't do two separate requests.
        if minlon < maxlon:
            filters = ['%s>=%s' % (lon.id, minlon), '%s<=%s' % (lon.id, maxlon),
                       '%s>=%s' % (lat.id, minlat), '%s<=%s' % (lat.id, maxlat)]
            constraints += '&' + '&'.join(filters)

    # Re-request dataset with the new constraints.
    if seqs: dataset = self._parseconstraints(constraints)

    headers = [('Content-description', 'dods_wms'),
               ('XDODS-Server', 'dods/%s' % '.'.join([str(i) for i in __dap__])),
              ]

    format = query.get('FORMAT', 'image/png')
    headers.append(('Content-type', format))
    format = format.split('/')[1]

    im = _dispatch(dataset, query)
    buf = StringIO()
    im.save(buf, format)
    output = buf.getvalue()

    return headers, output


def _capabilities(dapvar, description, environ):
    headers = [('Content-type', 'text/xml'),
               ('XDODS-Server', 'dods/%s' % '.'.join([str(i) for i in __dap__])),
              ]

    xml = """<?xml version='1.0' encoding="UTF-8" standalone="no" ?>
<WMT_MS_Capabilities version="1.1.1" xmlns:t="http://johnnydebris.net/xmlns/templess">

<Service>
  <Name t:content="name" />
  <Title t:content="title" />
  <OnlineResource xmlns:xlink="http://www.w3.org/1999/xlink" t:attr="xlink:href location" />
</Service>

<Capability>
  <Request>
    <GetCapabilities>
      <Format>application/vnd.ogc.wms_xml</Format>
      <DCPType>
        <HTTP>
          <Get><OnlineResource xmlns:xlink="http://www.w3.org/1999/xlink" t:attr="xlink:href location" /></Get>
        </HTTP>
      </DCPType>
    </GetCapabilities>
    <GetMap>
      <Format>image/jpeg</Format>
      <Format>image/png</Format>
      <Format>image/gif</Format>
      <Format>image/tiff</Format>
      <DCPType>
        <HTTP>
          <Get><OnlineResource xmlns:xlink="http://www.w3.org/1999/xlink" t:attr="xlink:href location" /></Get>
        </HTTP>
      </DCPType>
    </GetMap>
  </Request>
  <Exception>
    <Format>application/vnd.ogc.se_blank</Format>
  </Exception>
  <VendorSpecificCapabilities />
  <UserDefinedSymbolization SupportSLD="1" UserLayer="0" UserStyle="1" RemoteWFS="0"/>
  <Layer>
    <Title t:content="title" />
    <SRS>EPSG:4326</SRS>
    <LatLonBoundingBox t:attr="minx minx; miny miny; maxx maxx; maxy maxy" />
    <Layer t:content="layer">
      <Name t:content="name" />
      <Title t:content="title" />
      <Abstract t:content="abstract" />
    </Layer>
  </Layer>
</Capability>
</WMT_MS_Capabilities>"""

    # Get lat/lon range for the dataset, if available.
    try: lon_range = dapvar.attributes['NC_GLOBAL']['lon_range']
    except: lon_range = -180.0, 180.0
    try: lat_range = dapvar.attributes['NC_GLOBAL']['lat_range']
    except: lat_range = -90.0, 90.0

    # Build location using query_string but removing WMS variables.
    # We assume the request is done at the url::
    #
    #     http://server/dataset.wms?var1,var2,var3&CE&REQUEST=GetCapabilities
    #     ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ ^^^^^^^^^^^^^^^^^^^^^^^
    #                        DODS                               WMS
    #
    # ie, with a DODS part and a WMS part.
    location = construct_url(environ, with_query_string=True)
    location = re.sub("(&|\?)REQUEST=[^&]*", "", location)

    context = {'name': dapvar.name,
               'title': description,
               'location': location,
               'minx': str(lon_range[0]),
               'maxx': str(lon_range[1]),
               'miny': str(lat_range[0]),
               'maxy': str(lat_range[1]),
               'layer': [{'name': var.id,
                          'title': getattr(var, 'long_name', var.name),
                          'abstract': getattr(var, 'history', '')}
                         for var in dapvar.walk() if _islayer(var)]}
    t = templess.template(xml)
    output = ["""<?xml version='1.0' encoding="ISO-8859-1" standalone="no" ?>
<!DOCTYPE WMT_MS_Capabilities SYSTEM "http://schemas.opengis.net/wms/1.1.1/WMS_MS_Capabilities.dtd"
 [
 <!ELEMENT VendorSpecificCapabilities EMPTY>
 ]>
<!-- end of DOCTYPE declaration -->

""", t.unicode(context).encode('utf-8')]

    return headers, output


def _islayer(dapvar):
    if isinstance(dapvar, GridType):
        # Grids need lat and lon maps with coards units.
        lat = lon = None
        for var in dapvar.maps.values():
            if re.match("degrees?_n", var.units, re.IGNORECASE): lat = True
            elif re.match("degrees?_e", var.units, re.IGNORECASE): lon = True
        return lat and lon
    elif isinstance(dapvar, SequenceType):
        # Sequences also need lat and lon, following dapper spec.
        lat = lon = None
        for var in dapvar.walk():
            if getattr(var, "axis", None) == 'X': lon = True
            elif getattr(var, "axis", None) == 'Y': lat = True
        return lat and lon
    elif isinstance(dapvar, StructureType):
        # Structures need at least a valid variable.
        for var in dapvar.walk():
            if _islayer(var): return True
    
    return False


def _dispatch(dapvar, query):
    func = {DatasetType  : _dataset,
            StructureType: _structure,
            SequenceType : _sequence,
            BaseType     : _base,
            ArrayType    : _array,
            GridType     : _grid,
           }[type(dapvar)]

    return func(dapvar, query)


def _dataset(dapvar, query):
    """
    Join images together.
    """
    # Build base image and paste other over it.
    size = int(query.get('WIDTH', 256)), int(query.get('HEIGHT', 256))
    layers = query.get('LAYERS', '').split(',')

    # Transparent image? We need to set only the base image, all
    # others are transparent where there's no data.
    transparent = query.get('TRANSPARENT', 'FALSE')
    if transparent == 'TRUE': alpha = 0
    else: alpha = 255

    # Background color.
    bgcolor = query.get('BGCOLOR', '0xFFFFFF').replace('0x', '#')
    red, green, blue = ImageColor.getrgb(bgcolor)
    
    im = Image.new('RGBA', size, (red, green, blue, alpha))
    for var in dapvar.walk():
        if var.id in layers:
            try:
                layer = _dispatch(var, query)
                if layer: im.paste(layer, None, layer)
            except:
                pass
    return im


def _structure(dapvar, query):
    im = Image.new('RGBA', size, (0, 0, 0, 0))
    for var in dapvar.walk():
        if var.id in layers:
            try:
                layer = _dispatch(var, query)
                if layer: im.paste(layer, None, layer)
            except:
                pass
    return im


def _grid(dapvar, query):
    # Get lat & lon.
    lat = lon = None
    for var in dapvar.maps.values():
        if re.match("degrees?_n", var.units, re.IGNORECASE):
            lat = numpy.array(dapvar.maps[var.name][:])
        elif re.match("degrees?_e", var.units, re.IGNORECASE):
            lon = numpy.array(dapvar.maps[var.name][:])

    # Make lon monotonic.
    for i in range(1, lon.shape[0]):
        if lon[i] < lon[i-1]: lon[i] += 360.

    # Get boundaries.
    bbox = [float(v) for v in query.get('BBOX', '-180,-90,180,90').split(',')]
    minlon, minlat, maxlon, maxlat = bbox

    # Create base image.
    width, height = size = int(query.get('WIDTH', 256)), int(query.get('HEIGHT', 256))
    im = Image.new('RGBA', size, (0, 0, 0, 0))

    # Draw image. We need to offset since lon can be in a different range.
    offset = numpy.ceil((lon[0] - maxlon)/360.0) * 360.0
    while (minlon + offset) < lon[-1]:
        x, y = lon - offset, lat

        i0 = bisect.bisect_right(x, minlon) - 1
        i1 = bisect.bisect_left(x, maxlon) + 1
        if y[1] > y[0]:
            j0 = bisect.bisect_right(y, minlat) - 1
            j1 = bisect.bisect_left(y, maxlat) + 1
        else:
            j0 = bisect.bisect_right(-y, -maxlat) - 1
            j1 = bisect.bisect_left(-y, -minlat) + 1
        i0 = max(0, i0)
        j0 = max(0, j0)

        # Get step. Since the image has at most width x height pixels, we can subsample
        # the data accordingly -- there's no need to retrieve data with higher resolution.
        dy = (maxlat - minlat) / height
        dlat = y[1] - y[0]
        jstep = max(int(dy/dlat), 1)

        dx = (maxlon - minlon) / width
        dlon = x[1] - x[0]
        istep = max(int(dx/dlon), 1)

        # Retrieve data.
        y = y[j0:j1:jstep]
        if i0 < i1:
            x = x[i0:i1:istep]
            data = dapvar[...,j0:j1:jstep,i0:i1:istep]
        else:
            # Cyclic boundary.
            x = numpy.concatenate((x[i0:], x[:i1]))
            data = numpy.concatenate((dapvar[...,j0:j1:jstep,i0::istep],
                                      dapvar[...,j0:j1:jstep,:i1:istep]), len(dapvar.shape)-1)

        data = numpy.array(data)
        data = data * getattr(dapvar, 'scale_factor', 1) + getattr(dapvar, 'add_offset', 0)
        data = ma.masked_equal(data, getattr(dapvar, 'missing_value', -1e33))
        mask = data.mask

        # Make data 2D and in range 0-255.
        while len(data.shape) > 2:
            data = ma.average(data, axis=0)
        if hasattr(dapvar, 'actual_range'):
            actual_range = dapvar.actual_range
        else:
            alldata = dapvar[:]
            if hasattr(dapvar, 'missing_value'):
                alldata = ma.masked_equal(alldata, dapvar.missing_value)
            actual_range = numpy.min(alldata), numpy.max(alldata)
        data = data - actual_range[0]
        data = data / (actual_range[1] - actual_range[0])
        data = data * 255.
        data = data.filled(255)
        data = data.astype('b')

        # Make mask 2D -- mask is lost with ma.average (remember to file bug).
        mask = mask.astype('b')
        mask.shape = (-1, mask.shape[-2], mask.shape[-1])
        mask = 1 - mask[0]
        ##mask = mask * 255.  # add some transparency 
        mask = mask * 200.
        mask = mask.astype('b')

        # Create image with gray palette.
        mix = numpy.concatenate((numpy.reshape(data, (-1, 1)), numpy.reshape(mask, (-1, 1))), 1)
        layer = Image.fromstring("LA", (data.shape[1], data.shape[0]), mix.tostring())

        # Put palette into colored layer. We need to convert 
        # it to mode "F" first, and then finally to "RGBA".
        clayer = layer.convert("L")
        clayer.putpalette(jet)
        layer = layer.convert("RGBA")
        layer.paste(clayer, None, layer)

        # Paste image. Here we map the image from the data boundaries,
        # defined by lon0, lon1, lat0, lat1 to the BBOX limits.
        lon0, lon1 = x[0], 2*x[-1] - x[-2]
        lat0, lat1 = y[0], 2*y[-1] - y[-2]
        x0 = layer.size[0] * (minlon - lon0) / (lon1 - lon0)
        x1 = layer.size[0] * (maxlon - lon0) / (lon1 - lon0)
        y0 = layer.size[1] * (maxlat - lat0) / (lat1 - lat0)
        y1 = layer.size[1] * (minlat - lat0) / (lat1 - lat0)
        layer = layer.transform(size, Image.EXTENT, (x0, y0, x1, y1))

        im.paste(layer, None, layer)
        offset += 360.0
    
    return im


def _sequence(dapvar, query):
    # Find lat & lon.
    lat = lon = None
    for i, var in enumerate(dapvar.walk()):
        if getattr(var, "axis", None) == 'X': lon = i
        elif getattr(var, "axis", None) == 'Y': lat = i

    # Get boundaries.
    bbox = [float(v) for v in query.get('BBOX', '-180,-90,180,90').split(',')]
    minlon, minlat, maxlon, maxlat = bbox

    # Create image.
    width, height = size = int(query.get('WIDTH', 256)), int(query.get('HEIGHT', 256))
    im = Image.new('RGBA', size, (0, 0, 0, 0))
    draw = ImageDraw.Draw(im)

    # Draw point for each record. Since the bbox can be much larger than
    # the lon range (say, -360 to 360) we create an offset and change the
    # lon values at 360 degress increments.
    data = list(dapvar.data)
    offset = numpy.ceil((query['lon_range'][1] - maxlon)/360.0) * 360.0
    while (minlon + offset) < query['lon_range'][1]:
        for values in data:
            x, y = values[lon] - offset, values[lat]
        
            # Convert coords.
            x = width * (x - minlon) / (maxlon - minlon)
            y = height * (1 - (y - minlat) / (maxlat - minlat))
            draw.ellipse((x-2, y-2, x+2, y+2), fill=(255, 0, 0), outline=(100, 0, 0))

        offset += 360.0

    del draw
    return im
        

def _array(dapvar, query):
    pass

_base = _array



from __future__ import division

__author__ = 'Roberto De Almeida <rob@pydap.org>'

from paste.request import construct_url, parse_dict_querystring

from dap.lib import __dap__
from dap.responses.wms import templess, _islayer


xml = """<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://earth.google.com/kml/2.0" xmlns:t="http://johnnydebris.net/xmlns/templess">
<Document>
<name t:content="name" />
<description t:content="description" />
<visibility>1</visibility>
<open>1</open>
<NetworkLink t:content="layer">
  <name t:content="name" />
  <flyToView>0</flyToView>
  <visibility>0</visibility>
  <Url>
    <href t:content="location" />
    <viewRefreshMode>onStop</viewRefreshMode>
    <viewRefreshTime>1</viewRefreshTime>
    <ViewFormat>BBOX=[bboxWest],[bboxSouth],[bboxEast],[bboxNorth]</ViewFormat>
  </Url>
<refreshVisibility>1</refreshVisibility>
</NetworkLink>
</Document>
</kml>"""

overlay = """<?xml version="1.0" encoding="UTF-8"?>
<GroundOverlay xmlns:t="http://johnnydebris.net/xmlns/templess">
  <Name t:content="name" />
  <Icon>
    <href t:content="location" />
  </Icon>
  <LatLonBox>
    <north t:content="north" />
    <south t:content="south" />
    <east t:content="east" />
    <west t:content="west" />
  </LatLonBox>
</GroundOverlay>"""
    
    
def build(self, constraints=None):
    # Parse query string.
    query = parse_dict_querystring(self.environ)
    layers = query.get('LAYERS')
    location = construct_url(self.environ, with_query_string=True)

    # Get dataset.
    dataset = self._parseconstraints(constraints)

    headers = [('Content-type', 'application/keyhole'),
               ('XDODS-Server', 'dods/%s' % '.'.join([str(i) for i in __dap__])),
              ]

    # "Reflector" for Google Earth
    # If LAYERS is not specified GE is requesting the "static" KML. This KML
    # described the NetworkLinks, pointing back to this location. GE will
    # connect to these NetworkLinks passing a LAYERS var and possibly a BBOX;
    # in this case we return a ground overlay pointing to the WMS server.
    if layers:
        # Return overlay. Although LAYERS can request more than one layer,
        # the default behaviour for GE is requesting a single layer due to 
        # the KML file.
        location = location.replace('.kml', '.wms')
        location += '&SERVICE=WMS&VERSION=1.1.1&REQUEST=GetMap&SRS=EPSG:4326&WIDTH=512&HEIGHT=512&TRANSPARENT=TRUE&FORMAT=image/png'
        location = templess.cdatanode(location, None)
        bbox = query.get('BBOX', '-180,-90,180,90')
        bbox = bbox.split(',')
        context = {'name': layers,
                   'location': location,
                   'north': bbox[3],
                   'south': bbox[1],
                   'east': bbox[2],
                   'west': bbox[0]}
        t = templess.template(overlay)
    else:
        # Return a NetworkLink element for each layer, pointing back to this
        # file with a LAYERS variable.
        layers = []
        for var in dataset.walk():
            if _islayer(var):
                c = {'name': getattr(var, 'long_name', var.name),
                     'location': '%s?LAYERS=%s' % (location, var.id)}
                layers.append(c)
        context = {'description': self.description,
                   'name': dataset.name,
                   'layer': layers}
        t = templess.template(xml)

    output = ["""<?xml version="1.0" encoding="UTF-8"?>\n""", t.unicode(context).encode('utf-8')]

    return headers, output



# If true, then the svn revision won't be used to calculate the
# revision (set to True for real releases)
RELEASE = True

from setuptools import setup, find_packages
import sys, os

classifiers = """\
Development Status :: 4 - Beta
Environment :: Console
Intended Audience :: Developers
Intended Audience :: Science/Research
License :: OSI Approved :: MIT License
Operating System :: OS Independent
Programming Language :: Python
Topic :: Internet
Topic :: Scientific/Engineering
Topic :: Software Development :: Libraries :: Python Modules
"""

version = '0.3'

setup(name='dap.responses.wms',
      version=version,
      description="WMS response for pydap server",
      long_description="""\
pydap--WMS bridge.

The latest version is available in a `Subversion repository
<http://pydap.googlecode.com/svn/trunk/responses/wms#egg=dap.responses.wms-dev>`_.""",
      classifiers=filter(None, classifiers.split("\n")),
      keywords='wms dap opendap dods data',
      author='Roberto De Almeida',
      author_email='rob@pydap.org',
      url='http://pydap.org/responses/wms',
      license='MIT',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      namespace_packages=['dap.responses'],
      include_package_data=True,
      zip_safe=True,
      install_requires=[
          # -*- Extra requirements: -*-
          #'PIL',
          'dap[server]>=2.2.5.1',
          'Paste',
      ],
      entry_points="""
      # -*- Entry points: -*-
      [dap.response]
      wms = dap.responses.wms
      kml = dap.responses.wms.kml
      """,
      )
      
